from __future__ import annotations

import importlib.resources
import os
from pathlib import Path
from typing import Literal

from fastapi import FastAPI, Request, Response, status
from fastapi.responses import HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from pydantic import BaseModel, Field
from starlette.templating import _TemplateResponse  # noqa: TCH002
from uvicorn import Config, Server

from . import __version__
from .actions import store_heartbeat_from_fitbit
from .graphs import get_heartbeat_graph


class SpaceActionEvent(BaseModel):
    event_id: Literal["store_data_from_fitbit"] = Field(alias="id")
    trigger: str


class SpaceAction(BaseModel):
    event: SpaceActionEvent


resource_root_path = Path(str(importlib.resources.files("heartbeat_status")))
app = FastAPI()
templates = Jinja2Templates(directory=resource_root_path / "templates")

app.mount("/static", StaticFiles(directory=resource_root_path / "static_files"))


@app.get("/fitbit/callback")
async def fitbit_callback(code: str, state: str) -> RedirectResponse:  # noqa: ARG001
    return RedirectResponse(url="/")


@app.api_route(
    "/healthcheck",
    methods=["GET", "HEAD"],
    status_code=status.HTTP_200_OK,
)
async def healthcheck() -> Response:
    return Response(content="OK", media_type="text/plain")


@app.api_route("/", methods=["GET", "HEAD"], response_class=HTMLResponse)
async def root(request: Request) -> _TemplateResponse:
    return templates.TemplateResponse(
        "index.j2",
        {"request": request, "version": __version__},
    )


@app.get("/graph/heartbeat")
async def heartbeat_graph() -> HTMLResponse:
    return HTMLResponse(content=get_heartbeat_graph(), status_code=200)


@app.api_route(
    "/__space/v0/actions",
    methods=["POST", "HEAD"],
)
async def space_actions(action: SpaceAction) -> Response:
    if action.event.event_id == "store_data_from_fitbit":
        store_heartbeat_from_fitbit()
        return Response(content="OK", media_type="text/plain")
    return Response(content="NG", media_type="text/plain", status_code=status.HTTP_405_METHOD_NOT_ALLOWED)


def main() -> None:
    config = Config(
        app,
        port=5000,
        log_level="info",
        reload=bool(os.environ.get("DEBUG")),
    )
    server = Server(config)
    server.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

from deta import Deta  # type: ignore[import-untyped]
from pydantic import BaseModel

__DETA_BASE_CREDENTIAL_DB_KEY = "fitbit_credential_db"


class DetaBaseItem(BaseModel):
    key: str
    value: str


def get_deta_base_val(key: str) -> str | None:
    res = Deta().Base(__DETA_BASE_CREDENTIAL_DB_KEY).get(key)
    if not res:
        return None
    item = DetaBaseItem.model_validate(res)
    if isinstance(item.value, str) and len(item.value) > 0:
        return item.value
    return None


def set_deta_base_val(key: str, val: str) -> None:
    item = DetaBaseItem.model_validate(
        Deta().Base(__DETA_BASE_CREDENTIAL_DB_KEY).put({"value": val}, key),
    )
    if item.key == key and item.value == val:
        return

    msg = f"sent <{key}: {val}> to Deta Base, but got <{item.key}: {item.value}>"
    raise ValueError(msg)



from __future__ import annotations

from datetime import datetime, timedelta, timezone

import pandas as pd
import plotly.express as px  # type: ignore[import-untyped]

from .deta import get_deta_base_val
from .models import HeartBeatDataByMin

JST = timezone(timedelta(hours=+9), "JST")


def get_heartbeat_graph() -> str:
    l_data: list[tuple[datetime, int]] = []
    now = datetime.now(JST)
    date_keys = [(now + timedelta(hours=-i)).strftime("%Y-%m-%d") for i in range(7)]
    for date_key in date_keys:
        val = get_deta_base_val(date_key)
        if not val:
            continue
        data = HeartBeatDataByMin.model_validate_json(val).root
        l_data.extend(
            (
                datetime.fromisoformat(f"{date_key}T{dataset.time}"),
                dataset.value,
            )
            for dataset in data
        )
    data_frame = pd.DataFrame(l_data, columns=["Datetime", "BPM"]).set_index("Datetime")
    return str(px.line(data_frame=data_frame).to_html())



from __future__ import annotations

import os

import requests
from dotenv import find_dotenv, load_dotenv
from pydantic import BaseModel

from .deta import get_deta_base_val, set_deta_base_val


class FitbitAPIRefreshedResponse(BaseModel):
    access_token: str
    expires_in: int
    refresh_token: str
    scope: str
    token_type: str
    user_id: str


class FitbitAPICredential(BaseModel):
    FITBIT_CLIENT_ID: str
    FITBIT_CLIENT_SECRET: str
    FITBIT_ACCESS_TOKEN: str
    FITBIT_REFRESH_TOKEN: str


def refresh_credential() -> FitbitAPICredential:
    cred = __get_fitbit_credential()
    res = requests.post(
        "https://api.fitbit.com/oauth2/token",
        data={
            "grant_type": "refresh_token",
            "client_id": cred.FITBIT_CLIENT_ID,
            "refresh_token": cred.FITBIT_REFRESH_TOKEN,
        },
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        timeout=10,
    )
    token = FitbitAPIRefreshedResponse.model_validate(res.json())
    access_token, refresh_token = token.access_token, token.refresh_token
    set_deta_base_val("FITBIT_ACCESS_TOKEN", access_token)
    set_deta_base_val("FITBIT_REFRESH_TOKEN", refresh_token)
    return __get_fitbit_credential()


def __get_fitbit_credential() -> FitbitAPICredential:
    def _get_val(key: str, *, is_deta_base: bool = False) -> str:
        val = get_deta_base_val(key) if is_deta_base else None
        if not is_deta_base or not val:
            val = os.environ.get(key)
        if isinstance(val, str) and len(val) > 0:
            return val

        msg = f"Failed to fetch item -- {key}"
        raise ValueError(msg)

    load_dotenv(find_dotenv(), verbose=True)
    return FitbitAPICredential(
        FITBIT_CLIENT_ID=_get_val("FITBIT_CLIENT_ID"),
        FITBIT_CLIENT_SECRET=_get_val("FITBIT_CLIENT_SECRET"),
        FITBIT_ACCESS_TOKEN=_get_val("FITBIT_ACCESS_TOKEN", is_deta_base=True),
        FITBIT_REFRESH_TOKEN=_get_val("FITBIT_REFRESH_TOKEN", is_deta_base=True),
    )



from __future__ import annotations

import json
from datetime import datetime, timedelta, timezone

import requests

from .deta import set_deta_base_val
from .fitbit import refresh_credential
from .models import TodayHeartBeatByMin

JST = timezone(timedelta(hours=+9), "JST")


def store_heartbeat_from_fitbit() -> None:
    cred = refresh_credential()
    now = datetime.now(JST)
    yesterday = now - timedelta(days=1)
    for i in (now, yesterday) if now.hour == 0 else (now,):
        url = f"https://api.fitbit.com/1/user/-/activities/heart/date/{i.strftime('%Y-%m-%d')}/1d/1sec.json"
        heart = TodayHeartBeatByMin.model_validate(
            requests.get(
                url,
                headers={"Authorization": f"Bearer {cred.FITBIT_ACCESS_TOKEN}"},
                timeout=10,
            ).json(),
        )
        key = heart.activities_heart[0].date_time
        val = heart.activities_heart_intraday.dataset
        set_deta_base_val(key, json.dumps([d.model_dump() for d in val]))



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field, RootModel


# activities_heart
class _Value__HeartRateZone(BaseModel):  # noqa: N801
    calories_out: float = Field(alias="caloriesOut")
    rate_max: int = Field(alias="max")
    rate_min: int = Field(alias="min")
    minutes: int
    name: str


class _ActivitiesHeart__Value(BaseModel):  # noqa: N801
    custom_heart_rate_zones: list[Any] = Field(alias="customHeartRateZones")  # unconfirmed
    heart_rate_zones: list[_Value__HeartRateZone] = Field(alias="heartRateZones")
    resting_heart_rate: int = Field(alias="restingHeartRate")


class _TodayHeartBeatByMin__ActivitiesHeart(BaseModel):  # noqa: N801
    date_time: str = Field(alias="dateTime")
    value: _ActivitiesHeart__Value


# activities_heart_intraday
class _ActivitiesHeartIntraday__Dataset(BaseModel):  # noqa: N801
    time: str
    value: int


_ActivitiesHeartIntraday__Datasets = list[_ActivitiesHeartIntraday__Dataset]


class _TodayHeartBeatByMin__ActivitiesHeartIntraday(BaseModel):  # noqa: N801
    dataset: _ActivitiesHeartIntraday__Datasets
    dataset_interval: int = Field(alias="datasetInterval")
    dataset_type: str = Field(alias="datasetType")


# root
class TodayHeartBeatByMin(BaseModel):
    activities_heart: list[_TodayHeartBeatByMin__ActivitiesHeart] = Field(alias="activities-heart")
    activities_heart_intraday: _TodayHeartBeatByMin__ActivitiesHeartIntraday = Field(alias="activities-heart-intraday")


HeartBeatDataByMin = RootModel[_ActivitiesHeartIntraday__Datasets]
